//
// KMaplay's main widget implementation
//
// (c) 1997 Bjarni Ivarsson
//
	
#include "kmpg.h"
#include <kiconloader.h>

#define VERSION "0.3.2"

KMaplay::KMaplay( QWidget *parent, const char *name ) :
   QWidget( parent, name )
{
   player = new Player( );
   timer = new QTimer( this );
   prefsdialog = new Preferences( NULL, player );
   playlist = new Playlist(prefsdialog);
   filedialog = NULL;
   dropZone = new KDNDDropZone( this, DndURL );
   connect( dropZone, SIGNAL( dropAction( KDNDDropZone* )),
	    SLOT( onDrop( KDNDDropZone*)));

   status = Nofile;
   drawPanel();
   loadBitmaps();

   connect( playlist, SIGNAL(fileSelected(const char*)), SLOT(openFile(const char*)) );
   connect( timer, SIGNAL(timeout()), SLOT(onTimer()) );
   connect( slider, SIGNAL(valueChanged(int)),SLOT(newPos(int)) );
   connect( playPB, SIGNAL(clicked()), SLOT(playClicked()) );
   connect( pausePB, SIGNAL(clicked()), SLOT(pauseClicked()) );
   connect( stopPB, SIGNAL(clicked()), SLOT(stopClicked()) );
   connect( prevPB, SIGNAL(clicked()), playlist, SLOT(selectPrev()) );
   connect( nextPB, SIGNAL(clicked()), playlist, SLOT(selectNext()) );
   connect( fwdPB, SIGNAL(clicked()), SLOT(fwdClicked()) );
   connect( bwdPB, SIGNAL(clicked()), SLOT(bwdClicked()) );

   if (qApp->argc()==2)
      openFile( qApp->argv()[1] );
}

KMaplay::~KMaplay()
{
//   cerr << "Exiting" << endl;
   if ( status==Playing )
      player->stop(); // Make sure the child process exits!
}

void KMaplay::closeEvent ( QCloseEvent *e ) 
{
   prefsdialog->applyPrefs();
   e->accept();
}

// Create, place and set the name of a button
QPushButton *KMaplay::makeButton( int x, int y, int w, int h, const char *n )
{
   QPushButton *pb = new QPushButton( n, this );
   pb->setGeometry( x, y, w, h );
   return pb;
}

void KMaplay::onTimer()
{
   int sec;

   current_frame=player->loop();
   slider->setValue(current_frame);
   sec=(int)(current_frame*sinfo.ms_per_frame/1000);

   if (sec!=current_sec)
   {
      ledPanel->setTime( sec );
      current_sec=sec;
   }

  if(current_frame==-1)
  {
    stopClicked();
    playlist->selectNext();
  }
}
		
void KMaplay::newPos(int _pos)
{
   player->seek(_pos);
}

// Widgets have to be placed by manually......
void KMaplay::drawPanel()
{
   int ix = 0;
   int iy = 0;
   const int WIDTH = 30;
   const int HEIGHT = 32;

   setCaption( "MPEG Audio Player" );

   QPopupMenu *file = new QPopupMenu;
   CHECK_PTR( file );
   file->insertItem( "Open", this, SLOT(openClicked()), CTRL+Key_O );
   file->insertSeparator();
   file->insertItem( "Quit", this, SLOT(quitClicked()), ALT+Key_Q );

   QPopupMenu *view = new QPopupMenu;
   CHECK_PTR( view );
   view->insertItem( "Playlist Editor", playlist, SLOT(slotShow()), ALT+Key_E );
   view->insertItem( "Preferences", prefsdialog, SLOT(slotShow()) );

   QPopupMenu *help = new QPopupMenu;
   CHECK_PTR( help );
   help->insertItem( "Help Index", this, SLOT(slotHelp()), ALT+Key_H );
   help->insertSeparator();
   help->insertItem( "About", this, SLOT(aboutClicked()) );

   menu = new QMenuBar( this );
   CHECK_PTR( menu );
   menu->insertItem( "File", file );
   menu->insertItem( "View", view );
   menu->insertSeparator();
   menu->insertItem( "Help", help );

   ix = 3;
   iy = menu->frameGeometry().height();

   ledPanel = new LedPanel( this );
   ledPanel->move( 5, iy );

   iy += ledPanel->height();

   slider = new KSlider( KSlider::Horizontal, this );
   slider->setGeometry( ix, iy, WIDTH*7, 20 );
   slider->setRange( 0, 500 ); // 500 sec
   slider->setSteps( 15, 60 );

   iy += 22;
   ix=3;
   prevPB = makeButton( ix, iy, WIDTH, 24, "Prev" ); ix+=WIDTH;
   QToolTip::add( prevPB, "Previous track" );
   bwdPB = makeButton( ix, iy, WIDTH, 24, "Bwd" ); ix+=WIDTH;
   QToolTip::add( bwdPB, "15 secs backward" );
   pausePB = makeButton( ix, iy, WIDTH, 24, "Pause" ); ix+=WIDTH;
   pausePB->setToggleButton(TRUE);
   QToolTip::add( pausePB, "Pause" );
   playPB = makeButton( ix, iy, WIDTH, 24, "Play" ); ix+=WIDTH;
   playPB->setToggleButton(TRUE);	
   QToolTip::add( playPB, "Play" );
   stopPB = makeButton( ix, iy, WIDTH, 24, "Stop" ); ix+=WIDTH;
   QToolTip::add( stopPB, "Stop" );
   fwdPB = makeButton( ix, iy, WIDTH, 24, "Fwd" ); ix+=WIDTH;
   QToolTip::add( fwdPB, "15 secs forward" );
   nextPB = makeButton( ix, iy, WIDTH, 24, "Next" ); ix+=WIDTH*3/2;
   QToolTip::add( nextPB, "Next track" );

   iy += 28;

   resize( 216, iy);
   setFixedSize( 216, iy);

}

// Set up the bitmaps for the buttons
void KMaplay::loadBitmaps() 
{
  KIconLoader *loader = kapp->getIconLoader();

  pause_on  = loader->loadIcon( "pause_on.xpm" );
  pause_off = loader->loadIcon( "pause_off.xpm" );
  play_on   = loader->loadIcon( "play_on.xpm" );
  play_off  = loader->loadIcon( "play_off.xpm" );
  stop_on   = loader->loadIcon( "stop_on.xpm" );
  stop_off  = loader->loadIcon( "stop_off.xpm" );

  prevPM     = loader->loadIcon("prev.xpm");
  nextPM     = loader->loadIcon("next.xpm");
  prev_disPM = loader->loadIcon("prev.xpm");
  next_disPM = loader->loadIcon("next.xpm");

  QPixmap fwdPM(loader->loadIcon("ff.xpm") );
  QPixmap bwdPM(loader->loadIcon("rew.xpm") );

  playPB->setPixmap( play_off );
  pausePB->setPixmap( pause_off );                 
  stopPB->setPixmap( stop_on );
  prevPB->setPixmap( prev_disPM );
  nextPB->setPixmap( next_disPM );
  fwdPB->setPixmap( fwdPM );
  bwdPB->setPixmap( bwdPM );               
}

void KMaplay::openClicked()
{
   if (filedialog==NULL)
      filedialog = new QFileDialog( NULL, "*.mp?", this, "Open file", TRUE );
   filedialog->setCaption("Open file");

   if( filedialog->exec() == 1)
   {
      openFile( filedialog->selectedFile().data() );
   }
}

void KMaplay::updateButtons()
{
   if (status==Playing)
   {
      playPB->setOn(TRUE);
      pausePB->setOn(FALSE);
      playPB->setPixmap( play_on );
      pausePB->setPixmap( pause_off );
      stopPB->setPixmap( stop_off );
   }
   else if (status==Paused)
   {
      playPB->setOn(FALSE);
      pausePB->setOn(TRUE);
      playPB->setPixmap( play_off );
      pausePB->setPixmap( pause_on );
      stopPB->setPixmap( stop_off );
   }
   else if (status==Stopped || status==Nofile)
   {
      playPB->setOn(FALSE);
      pausePB->setOn(FALSE);
      playPB->setPixmap( play_off );
      pausePB->setPixmap( pause_off );
      stopPB->setPixmap( stop_on );
   }

   // MICHEL
   if(!playlist->isEmpty() )
   {
      prevPB->setPixmap( prevPM );
      nextPB->setPixmap( nextPM );
   } else {
      prevPB->setPixmap( prev_disPM );
      nextPB->setPixmap( next_disPM );
   }

   playPB->repaint(TRUE);
   pausePB->repaint(TRUE);
   stopPB->repaint(TRUE);
   nextPB->repaint(TRUE);
   prevPB->repaint(TRUE);
}

void KMaplay::playClicked()
{
   int retval;

   if (status!=Playing && status==Nofile)
      playlist->selectNext();


   if (status==Playing || status==Nofile)
   {
      updateButtons();
      return;
   }

   if ( (retval=player->play()) == PLAYER_OK )
   {
// 	  playlist->selectNext();
      timer->start(0);
      status=Playing;
      ledPanel->play();
      updateButtons();
      return;
   }

   if ( retval == PLAYER_DEVICEBUSY )
   {
      KMsgBox::message( this, "Error", "The audio device is busy\n",
			KMsgBox::STOP );
      updateButtons();
      return;
   }
   else if ( retval == PLAYER_DEVICEINVALID )
   {
      KMsgBox::message( this, "Error", "Unable to open audio device\n",
			KMsgBox::STOP );
      updateButtons();
      return;
   }
}

void KMaplay::pauseClicked()
{
   if (status==Stopped || status==Nofile)
   {
      updateButtons();
      return;
   }
   if (status==Paused)
   {
      playClicked();
   }
   else
   {
      timer->stop();
      player->pause();
      status=Paused;
      ledPanel->pause();
      updateButtons();
   }
}

void KMaplay::stopClicked()
{
   if (status==Nofile)
      return;

   if (status==Paused)
   {
      //    blinker->stop();
      //setLCDColor( trackTimeLCD, green, black );
   }
   else if (status==Playing)
   {
      timer->stop();
   }
   status=Stopped;
   ledPanel->stop();
   current_frame=0;
   player->stop();
   slider->setValue(0);
   updateButtons();
}

void KMaplay::prevClicked()
{
}

void KMaplay::nextClicked()
{
}

void KMaplay::fwdClicked()
{
   if(status==Playing)
      player->seek(current_frame+ffrw_frames);
}

void KMaplay::bwdClicked()
{
   if(status==Playing)
      player->seek(current_frame-ffrw_frames);
}

void KMaplay::openFile(const char *_filename )
{
   int retval;

   stopClicked();
   player->closeFile();

   if ((retval=player->openFile( _filename, &sinfo)) < 0)
   {
      if (retval==-1)
	 KMsgBox::message( this, "Error", "The selected file is not valid\n"\
			   "and cannot be played by this version of KMaplay",
			   KMsgBox::STOP );
      else
	 KMsgBox::message( this, "Error", "The selected file cannot be opened\n",
			   KMsgBox::STOP );
      return;
   }

   ledPanel->setInfo( &sinfo );
   ffrw_frames = (int)(15000/sinfo.ms_per_frame); // 15 sec
   slider->setRange( 0, sinfo.min_number_of_frames );

   int st1=(int)(15000/sinfo.ms_per_frame), st2=(int)(60000/sinfo.ms_per_frame);
   float len=sinfo.min_number_of_frames*sinfo.ms_per_frame;

   total_sec=(int)(len/1000);

   if (len>=360000)
   {
      st1 = st1*2;
   }
   while (len>=480000)
   {
      st1 *= 2;
      st2 = st1;
      len /= 2;
   }

   slider->setValue(0);
   slider->setSteps( st1, st2 );

   current_frame=0;
   current_sec=-1;
   status = Stopped;
   
   playClicked();
}

void KMaplay::quitClicked()
{
   bool ok = TRUE;

   if (ok)
   {
      timer->stop();
      qApp->quit();
   }
}

void KMaplay::aboutClicked()
{
   KMsgBox::message( this, "About kmpg " VERSION ,
		     "(c) 1997 Bjarni Ivarsson (a94bjaiv@ida.his.se)\n\n"\
		     "See Help:Index for more information",
		     KMsgBox::EXCLAMATION);
}

void KMaplay::onDrop( KDNDDropZone* _zone )
{
   QStrList strlist;
   KURL *url;
   char *str;

   // cerr << "URLs dropped on kmaplay:" << endl;
   strlist = _zone->getURLList();
   // for( str = strlist.first(); str != 0; str=strlist.next() )
   //   cerr << str << endl;

   if ( strlist.count() >= 1 )
   {
      playlist->clear();
      for( str=strlist.first(); str!=0; str=strlist.next() )
      {
	 url = new KURL( str );
	 playlist->addFile( url->path() );
	 delete url;
      }
      playlist->selectNext();
   }
   /*
   else
   {
      // Open first URL
      url = new KURL( strlist.first() );
      openFile( url->path() );
      delete url;
   }
   */
}

void KMaplay::slotHelp()
{
   KApplication::getKApplication()->invokeHTMLHelp( 
      "kmpg/kmpg.html", "" );
}




