#include <qpainter.h>
#include <qregexp.h>
#include <kapp.h>
#include <kiconloader.h>

#include "ledpanel.h"

LedPanel::LedPanel( QWidget *parent, const char *name ):
   QWidget( parent, name )
{
   setFixedSize( 206, 58 );
   pm_back.resize( 206, 70 );
   pm_back.fill(QColor(0,0,0));
   loadPixmaps();
   bitBlt(&pm_back, 0, 0, pm_panel );
   setBackgroundPixmap( pm_back );
   for( int i=0; i<5; i++ ) 
      value_digits[i] = -1;
   scrollTimer = new QTimer( this , "scrolltimer" );
   connect( scrollTimer, SIGNAL(timeout()), SLOT(scrollTimeout()) );
}

LedPanel::~LedPanel()
{
   delete scrollTimer;
   delete pm_panel;
}

void LedPanel::loadPixmaps()
{
   KIconLoader *loader = kapp->getIconLoader();
   QString digits[10] = 
   { "digit_0.xpm", "digit_1.xpm", "digit_2.xpm", "digit_3.xpm", 
     "digit_4.xpm", "digit_5.xpm", "digit_6.xpm", "digit_7.xpm", 
     "digit_8.xpm", "digit_9.xpm" };

   for( int i=0; i<10; i++ )
      pm_digits[i] = new QPixmap(  loader->loadIcon( digits[i] ) );
      pm_blank = new QPixmap(loader->loadIcon("digit_blank.xpm" ));
      pm_pause = new QPixmap(loader->loadIcon("digit_pause.xpm" ));
      pm_play = new QPixmap(loader->loadIcon("digit_play.xpm" ));
      pm_panel = new QPixmap(loader->loadIcon("panel.xpm" ));
}

void LedPanel::updateDigit( int _digit, int _value )
{
   QPixmap *pm;

   if( value_digits[_digit] == _value )
      return;
   
   if( _digit == 0 )
      switch( _value )
      {
	 case 1:
	    pm = pm_pause;
	    break;
	 case 2:
	    pm = pm_play;
	    break;
	 default:
	    pm = pm_blank;
	    break;
      }
   else
      if( _value>-1 && _value<10 )
	 pm = pm_digits[ _value ];
      else
	 pm = pm_blank;

   bitBlt(&pm_back, 18*_digit+(_digit>2?10:1), 2, pm );
   setBackgroundPixmap( pm_back );
}

void LedPanel::play()
{
   updateDigit( 0, 2 );
}

void LedPanel::pause()
{
   updateDigit( 0, 1 );
}

void LedPanel::stop()
{
   int i;
   
   for( i=0; i<5; i++ )
      updateDigit( i, 0 );
}

void LedPanel::setTime( int _secs )
{
   int mins, secs;
   
   mins = _secs/60;
   secs = _secs%60;

   updateDigit( 1, mins/10 );
   updateDigit( 2, mins%10 );
   updateDigit( 3, secs/10 );
   updateDigit( 4, secs%10 );
}

void LedPanel::setInfo( streaminfo *_sinfo )
{
   QPainter paint;
   QFont _font( "helvetica", 10 );
   QFontMetrics _fontmetrics( _font );
   QString string;
   int x;

   scrollTimer->stop();

   paint.begin( &pm_back );
   paint.setPen( QColor(0,255,0) );
   paint.setFont( _font );
   paint.fillRect(0,26,110,15,QBrush(QColor(0,0,0)));
   paint.fillRect(5,38,195,20,QBrush(QColor(0,0,0)));
   paint.fillRect(105,0,110,41,QBrush(QColor(0,0,0)));

   if( _sinfo->which_channels == 2 )
   {
      paint.setPen( QColor(255,0,0) );
      paint.drawText( 4, 36, "stereo" );
      paint.setPen( QColor(0,255,0) );
   }
   else
      paint.drawText( 4, 36, "mono" );
   
   if( _sinfo->original )
      paint.drawText( 40, 36, "original" );
   if( _sinfo->copyright )
      paint.drawText( 84, 36, "(c)" );
   
   string = "bitrate: ";
   string += _sinfo->bitrate_string;
   paint.drawText( 105, 12, string );
   string = "freq: ";
   string += _sinfo->freq_string;
   paint.drawText( 105, 24, string );
   paint.drawText( 105, 36, _sinfo->version_string );
   paint.drawText( 161, 36, _sinfo->layer_string );   
   
   paint.end();

   {
      int i;
      int sec = (int)(_sinfo->min_number_of_frames*_sinfo->ms_per_frame/1000);
      QRegExp r = QRegExp( ".mp[123]", FALSE, FALSE );
      QString s2;

      string = _sinfo->name;
      i = string.findRev( r );
      if( i!=-1)
	 string.remove(i, string.length());
      string += " [";
      s2.setNum(sec/60);
      string += s2;
      string += ":";
      if( (sec%60)<10 )
	 string += "0";
      s2.setNum(sec%60);
      string += s2;
      string += "]";
   }
   
   pm_filename.resize( _fontmetrics.width( string )+10, 20 );
   pm_filename.fill(QColor(0,0,0));
   paint.begin( &pm_filename );
   paint.setPen( QColor(0,255,0) );
   paint.setFont( _font );
   paint.drawText( 5, 8+_fontmetrics.height()/2, string );
   paint.end();
   bitBlt( &pm_back, 5, 38, &pm_filename, 0, 0, 195, 20 );

   setBackgroundPixmap( pm_back );
   scrollPos = 0;
   scrollInc = 4;
   if (pm_filename.width()>201)
      scrollTimer->start(200);
   else
      scrollTimeout();
}

void LedPanel::scrollTimeout()
{
   scrollPos += scrollInc;
   if( (scrollPos > pm_filename.width()-195) || (scrollPos==0) )
      scrollInc = -scrollInc;

   bitBlt( &pm_back, 5, 38, &pm_filename, scrollPos, 0, 195, 20 );
   setBackgroundPixmap( pm_back );
}

void LedPanel::paintEvent( QPaintEvent* )
{
}

void LedPanel::styleChange( GUIStyle )
{
}
