// 
// Heavily borrowed code from kdehelp :-)
//

#include "khelp.h"
#include "kmpg.h"

static void cleanup( int );
void catchSignals();

static int msgqid = -1;
static QString pidFile;

KMaplay *kmaplay;

//-----------------------------------------------------------------------------
// timer used to monitor msg queue for request for new window

class Timer : public QObject
{
public:
  Timer()
  {
    startTimer( 100 );
  }

protected:
  virtual void timerEvent( QTimerEvent * );
};

void Timer::timerEvent( QTimerEvent * )
{
  KHelpMsg buf, retbuf;
  
  if ( buf.recv( msgqid, 1L ) != -1 )
    {
//      printf( "got request: %s\n", buf.getMsg() );
      
      retbuf.setType( 2L );
      retbuf.send( msgqid );
      
      QString url = buf.getMsg();
      
      if ( !strchr( url, ':' ) )
	{
	  url = "";
	  url += buf.getMsg();
	}
      kmaplay->openFile((const char*) url.data() );
    }
}

int main ( int argc, char *argv[] )
{
  int i;
  QString url, initDoc;
  FILE *fp;

  srand(time(0)); //For the Shuffle Button

  if ( geteuid() == 0 ) {
  	pid_t mypid = getpid();
	setpriority(PRIO_PROCESS,mypid,-20);
	seteuid( getuid() );
  }
  
  for ( i = 1; i < argc; i++ ) {
    if ( argv[i][0] == '-' )
      continue;
    
    initDoc = argv[i];
    break;
  }

  if ( i == argc ) {
    initDoc = "";
  }

  // create data directory if necessary
  QString p = getenv( "HOME" );
  QString rcDir = p + "/.kmpg";
  if ( access( rcDir, F_OK ) )
    mkdir( rcDir, 0740 );
  
  pidFile = rcDir + "/kmpg.pid";

  // if there is a pidFile then this is not the first instance
  if ( ( fp = fopen( pidFile, "r" ) ) != NULL ) {
//    printf( "found PID file\n" );
    KHelpMsg buf;
    int pid;
    buf.setType( 1L );
    buf.setMsg( initDoc );
    fscanf( fp, "%d %d", &pid, &msgqid );
    // if this fails I assume that the pid file is left over from bad exit
    // and continue on
    //
    if ( buf.send( msgqid ) != -1) {
      // if we don't receive a reply within 1sec assume previous
      // instance died an unatural death.
      // How should this stuff be handled properly?
      //
      time_t start = time(NULL);
      while ( time(NULL) - start < 1 ) {
	if (buf.recv( msgqid, 2L, IPC_NOWAIT ) != -1) {
	  fclose( fp );
	  exit(0);
	}
      }
      msgctl( msgqid, IPC_RMID, NULL );
    }
    fclose( fp );
  }

  // This is the first instance so create a pid/msgqid file
  key_t key = ftok( argv[0], (char)rand() );
  msgqid = msgget( key, IPC_CREAT | 0600 );
//  cerr << argv[0] << endl;
  
  fp = fopen( pidFile, "w" );
//  fprintf( fp, "%d %d\n", getpid(), msgqid );
  fclose( fp );
  
  // so that everything is cleaned up
  //catchSignals();

  KApplication appkmaplay( argc, argv, "kmpg" );
  kmaplay = new KMaplay; 
//  Timer *timer = new Timer;

  appkmaplay.setMainWidget( kmaplay );
  kmaplay->show();
  appkmaplay.exec();

//  delete timer;
  
  msgctl( msgqid, IPC_RMID, 0 );
  
  remove( pidFile );
}

// make sure the pid file is cleaned up when exiting unexpectedly.
//
void catchSignals()
{
//      signal(SIGHUP, cleanup);                /* Hangup */
        signal(SIGINT, cleanup);                /* Interrupt */
        signal(SIGTERM, cleanup);               /* Terminate */
//      signal(SIGCHLD, cleanup);

        signal(SIGABRT, cleanup);
        signal(SIGALRM, cleanup);
        signal(SIGFPE, cleanup);
        signal(SIGILL, cleanup);
        signal(SIGPIPE, cleanup);
        signal(SIGQUIT, cleanup);
//      signal(SIGSEGV, cleanup);

#ifdef SIGBUS
        signal(SIGBUS, cleanup);
#endif
#ifdef SIGEMT
        signal(SIGEMT, cleanup);
#endif
#ifdef SIGPOLL
        signal(SIGPOLL, cleanup);
#endif
#ifdef SIGPROF
        signal(SIGPROF, cleanup);
#endif
#ifdef SIGSYS
        signal(SIGSYS, cleanup);
#endif
#ifdef SIGTRAP
        signal(SIGTRAP, cleanup);
#endif
#ifdef SIGVTALRM
        signal(SIGVTALRM, cleanup);
#endif

#ifdef SIGXCPU
        signal(SIGXCPU, cleanup);
#endif
#ifdef SIGXFSZ
        signal(SIGXFSZ, cleanup);
#endif
}

// remove pid file
static void cleanup( int sig )
{
   printf( "cleanup: signal = %d\n", sig );
   if (msgqid >= 0)
      msgctl( msgqid, IPC_RMID, 0 );
   remove( pidFile );
   delete kmaplay;
   exit(0);
}
