#include <qpushbt.h>
#include <qdir.h>
#include <qfile.h>
#include <qfileinf.h>
#include <qtstream.h>
#include "playlist.h"

Playlist::Playlist(Preferences *prefsdialog, QWidget *parent, const char *name ) :
  QDialog( parent, name )
{
   KDNDDropZone *dropZone = new KDNDDropZone( this, DndURL );
   connect( dropZone, SIGNAL( dropAction( KDNDDropZone* )),
	    SLOT( onDrop( KDNDDropZone*)));

  prefs=prefsdialog;
  listbox = new QListBox( this );
  setCaption( "Playlist Editor" );

  buttonPanel = new QWidget( this );

  QPushButton *pb1 = new QPushButton( "Close", buttonPanel );
  pb1->setGeometry( 10, 5, 80, 25 );
  QPushButton *pb2 = new QPushButton( "Load playlist", buttonPanel );
  pb2->setGeometry( 10, 40, 80, 25 );
  QPushButton *pb3 = new QPushButton( "Save playlist", buttonPanel );
  pb3->setGeometry( 10, 65, 80, 25 );
  QPushButton *pb4 = new QPushButton( "Add playlist", buttonPanel );
  pb4->setGeometry( 10, 100, 80, 25 );
  QPushButton *pb5 = new QPushButton( "Add MPEG", buttonPanel );
  pb5->setGeometry( 10, 125, 80, 25 );
  QPushButton *pb6 = new QPushButton( "up", buttonPanel );
  pb6->setGeometry( 10, 160, 80, 12 );
  QPushButton *pb7 = new QPushButton( "dn", buttonPanel );
  pb7->setGeometry( 10, 172, 80, 12 );
  QPushButton *pb8 = new QPushButton( "Del MPEG", buttonPanel );
  pb8->setGeometry( 10, 194, 80, 25 );
  QPushButton *pb9 = new QPushButton( "Clear list", buttonPanel );
  pb9->setGeometry( 10, 219, 80, 25 );
  QPushButton *pb10 = new QPushButton("Shuffle", buttonPanel);
  pb10->setGeometry( 10, 254, 80 ,25);

  QString base_path = KApplication::getKApplication()->localkdedir() + QString( "/share/apps/kmpg/pics/" );
  QPixmap pm_up( base_path + "up.xpm" );
  QPixmap pm_down( base_path + "down.xpm" );
  pb6->setPixmap( pm_up );
  pb7->setPixmap( pm_down );

  setMinimumSize( 350, 250 );
  connect( listbox, SIGNAL(selected(int)), SLOT(itemSelected(int)) );
  connect( pb1, SIGNAL(clicked()), SLOT(slotClose()) );
  connect( pb2, SIGNAL(clicked()), SLOT(slotLoadPlaylist()) );
  connect( pb3, SIGNAL(clicked()), SLOT(slotSavePlaylist()) );
  connect( pb4, SIGNAL(clicked()), SLOT(slotAddPlaylist()) );
  connect( pb5, SIGNAL(clicked()), SLOT(slotAddMPEG()) );
  connect( pb6, SIGNAL(clicked()), SLOT(slotUp()) );
  connect( pb7, SIGNAL(clicked()), SLOT(slotDown()) );
  connect( pb8, SIGNAL(clicked()), SLOT(slotDelMPEG()) );
  connect( pb9, SIGNAL(clicked()), SLOT(clear()) );
  connect( pb10, SIGNAL(clicked()), SLOT(slotShuffle()) );

  lastselected = -1;
}

// MICHEL
bool Playlist::isEmpty()
{
   if( listbox->count() ) {
      return false;
   } else {
      return true;
   }
}

void Playlist::onDrop(  KDNDDropZone* _zone )
{
	QStrList strlist;
	KURL *url;
	char *str;

  // cerr << "URLs dropped on playlist:" << endl;
   strlist = _zone->getURLList();
  // for( str=strlist.first(); str!=0; str=strlist.next() )
  //    cerr << str << endl;

   if ( strlist.count()>1 )
      for( str=strlist.first(); str!=0; str=strlist.next() )
      {
         url = new KURL( str );
         this->addFile( url->path() );
         delete url;
      }
   else
   {
      // Open first URL
      url = new KURL( strlist.first() );
      this->addFile( url->path() );
      delete url;
   }
}

void Playlist::resizeEvent( QResizeEvent* )
{
  listbox->setGeometry( 5, 5, width()-105, height()-10 );
  buttonPanel->setGeometry( width()-100, 0, 100, height() );
}

void Playlist::selectPrev()
{
  if ( listbox->count() == 0 )
    return;
  if ( lastselected == 0 )
    itemSelected( listbox->count()-1 );
  else
    itemSelected( lastselected-1 );
  listbox->setCurrentItem( lastselected );
}

void Playlist::selectNext()
{
  if ( listbox->count() == 0 )
    return;
  if ( listbox->count() > (lastselected+1) )
    itemSelected( lastselected+1 );
  else if ( prefs->getPlayEndless() ) {
    itemSelected( 0 );
    lastselected = 0;
    listbox->setCurrentItem( lastselected );
  } else
	return;
  listbox->setCurrentItem( lastselected );
}

void Playlist::itemSelected(int _item)
{
  emit fileSelected( (const char*)listbox->text(_item) );
  lastselected=_item;
}

void Playlist::addFile( const char *_filename )
{
  static int recursionDepth = 0;              // Guard against infinited
					      // recursion

  // show();

  QFileInfo fi(_filename);

  if (!fi.exists())
    return;

  if (recursionDepth > 1000) 
    return;

  recursionDepth++;

  if (fi.isDir()) {
    // Add all mp3 files in the directory

    QDir dir( _filename );
    dir.setNameFilter( "*.mp3" );

    const QFileInfoList *list = dir.entryInfoList();
    QFileInfoListIterator it( *list );
    QFileInfo *item;

    while ((item = it.current())) {        
      listbox->insertItem( dir.filePath( item->fileName() ) );
      ++it;   
    }
  }
  else if (fi.fileName().right(3) == "m3u") {
    // Add in the playlist

    QFile input( _filename );
    if (input.open( IO_ReadOnly )) {
      QTextStream stream( &input );

      do {
	QString string = stream.readLine();
	if (stream.eof()) 
	  break;

	addFile( string );
      } while (1);

      input.close();
    }
  }
  else if (fi.fileName().right(3) == "mp3") {
    // Add in the file
    listbox->insertItem( _filename );
  }

  recursionDepth--;
}

void Playlist::clear()
{
   listbox->clear();
//  hide();
   lastselected = -1;
}

void Playlist::slotShow()
{
   show();
}

void Playlist::slotClose()
{
   hide();
}

void Playlist::slotLoadPlaylist()
{
   QFile file;
   QTextStream stream;
   QString string, filename;

   QString dirname = KApplication::getKApplication()->localkdedir();
   QFileInfo fi(dirname + QString("/share/apps/kmpg/"));
   if ( !(fi.exists() && fi.isDir()) ) {
   		QDir dir = dirname.data();
		if( !dir.mkdir("share/apps/kmpg") ) 
			dirname = 0;
		else 
			dirname += "/share/apps/kmpg";
   } else
   		dirname += "/share/apps/kmpg/";

   filename = QFileDialog::getOpenFileName( dirname, "*.m3u" );
   if( filename.isNull() )
      return;
   
   file.setName( filename );
   file.open( IO_ReadOnly );
   stream.setDevice( &file );
   listbox->clear();
   string = stream.readLine();
   while( !stream.eof() )
   {
      listbox->insertItem( string );
      string = stream.readLine();
   }
   file.close();
}

void Playlist::slotSavePlaylist()
{
   QFile file;
   QTextStream stream;
   QString filename;

   QString dirname = KApplication::getKApplication()->localkdedir();
   QFileInfo fi(dirname + QString("/share/apps/kmpg/"));
   if ( !(fi.exists() && fi.isDir()) ) {
   		QDir dir = dirname.data();
		if( !dir.mkdir("share/apps/kmpg") ) 
			dirname = 0;
		else 
			dirname += "/share/apps/kmpg";
   } else
   		dirname += "/share/apps/kmpg/";

   filename = QFileDialog::getSaveFileName( dirname, "*.m3u" );
   if( filename.isNull() )
      return;
   
   file.setName( filename );
   file.open( IO_WriteOnly );
   stream.setDevice( &file );
   for( int i=0; i<listbox->count(); i++ )
   {
      stream << listbox->text(i) << endl;
   }
   file.close();
}

void Playlist::slotAddPlaylist()
{
   QFile file;
   QTextStream stream;
   QString string, filename;

   QString dirname = KApplication::getKApplication()->localkdedir();
   QFileInfo fi(dirname + QString("/share/apps/kmpg/"));
   if ( !(fi.exists() && fi.isDir()) ) {
   		QDir dir = dirname.data();
		if( !dir.mkdir("share/apps/kmpg") ) 
			dirname = 0;
		else 
			dirname += "/share/apps/kmpg";
   } else
   		dirname += "/share/apps/kmpg/";

   filename = QFileDialog::getOpenFileName( dirname, "*.m3u" );
   if( filename.isNull() )
      return;
   
   file.setName( filename );
   file.open( IO_ReadOnly );
   stream.setDevice( &file );
   string = stream.readLine();
   while( !stream.eof() )
   {
      listbox->insertItem( string );
      string = stream.readLine();
   }
   file.close();
}

void Playlist::slotAddMPEG()
{
   QString filename;

   filename = QFileDialog::getOpenFileName( 0, "*.mp?" );
   if( filename.isNull() )
      return;
   listbox->insertItem( filename );
}

void Playlist::slotUp()
{
   int _item = listbox->currentItem();
   
   if( _item < 1 )
      return;
   
   listbox->insertItem( listbox->text( _item ), _item-1 );
   listbox->removeItem( _item+1 );
   listbox->setCurrentItem( _item - 1 );
}

void Playlist::slotDown()
{
   int _item = listbox->currentItem();
   
   if( _item > listbox->count()-2 )
      return;

   listbox->insertItem( listbox->text( _item ), _item+2 );
   listbox->removeItem( _item );
   listbox->setCurrentItem( _item + 1 );
}

void Playlist::slotDelMPEG()
{
   if( listbox->currentItem() < 0 )
      return;
   
   listbox->removeItem( listbox->currentItem() );
}

void Playlist::slotShuffle()
{
  int i;
  QStrList randlist;

  listbox->setUpdatesEnabled (FALSE);  //disable screen updates
  while (listbox->count()) {
    i = rand() % listbox->count();  //best shuffle results with % !
    randlist.append(listbox->text(i));
    listbox->removeItem(i);
  }
  listbox->insertStrList(&randlist);
  listbox->setUpdatesEnabled (TRUE);  //enable screen updates again
  listbox->repaint ();  //update the changed list box
}
